/*  Smooth ND driver for Player/Stage
 *  
 *  SND Authors:  Joey Durham (algorithm) , 
 *                Luca Invernizzi (driver implementation)
 *
 *  Implemented on top of Player - One Hell of a Robot Server
 *  Copyright (C) 2003  (Brian Gerkey, Andrew Howard)
 *  
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/** @ingroup drivers */
/** @{ */
/** @defgroup driver_snd snd
 * @brief Smooth Nearness Diagram Navigation

This driver implements the Smooth Nearness Diagram Navigation algorithm.
This algorithm handles local reactive collision-avoidance and 
goal-seeking and is designed for non-holonomic robots operating in 
tight spaces.  The algorithm is in the following paper:

- J. W. Durham, F. Bullo. Smooth Nearness Diagram Navigation.  IEEE
International Conference on Intelligent Robots and Systems, p 690-695,
2008.  <a href="http://motion.mee.ucsb.edu/pdf/2008a-db.pdf">PDF</a>

SND is based off the Nearness Diagram Navigation method presented in:

- J. Minguez, L. Montano. Nearness Diagram Navigation (ND): Collision
Avoidance in Troublesome Scenarios. IEEE Transactions on Robotics and
Automation, pp 154, 2004.
<a href="http://webdiis.unizar.es/~jminguez/TRAND.pdf">PDF</a>

Briefly, like ND, SND is a gap or discontinuity-based method.  The main
differences are that SND uses a single motion law instead of picking
between six, and SND considers all close obstacle points instead of only the
closest point on each side.  Together these two changes reduce the jerky
turns which can occur with ND, producing smoother motion.

NOTE:  Like ND, SND is a local, reactive path planner.  There are
limits to what it can do on its own.  It can successfully move through
tight, cluttered environments with local traps, but will often get
stuck in larger traps or at ambiguous branch points if asked to 
navigate long distances through complex environments.  It infers local
topological information from laser scans but does not know or remember
global topological information.  For large, complex environments, SND
should be fed a series of waypoints to avoid larger traps.  It could be
combined with an implementation of D* using a coarse, static
representation of the environment.

This driver reads pose information from a @ref interface_position2d
device, sensor data from a @ref interface_laser device, and writes 
commands to a @ref interface_position2d device.  The two 
@ref interface_position2d devices can be the same.

The driver itself supports the @interface_position2d interface.  Send
@ref PLAYER_POSITION2D_CMD_POS commands to set the goal pose.  The driver
also accepts @ref PLAYER_POSITION2D_CMD_VEL commands, simply passing them
through to the underlying output device.


@par Compile-time dependencies

- none

@par Provides

- @ref interface_position2d

@par Requires

- "input" @ref interface_position2d : source of pose and velocity information
- "output" @ref interface_position2d : sink for velocity commands to control the robot
- @ref interface_laser : the laser to read from

@par Configuration requests

- all @ref interface_position2d requests are passed through to the
underlying "output" @ref interface_position2d device.

@par Configuration file options

- robot_radius (length)
  - Default: 0.25 m
  - Radius of the robot.  In the driver the robot is assumed to be a 
    circle of this radius centered around the laser sensor.  In 
    practice, navigation is quite forgiving of rectangular shapes and
    off-center sensors so long as the tightness of squeezes between 
    clutter in the environment loosely matches the accuracy of the 
    circular approximation.  The robot will stop moving if it believes
    there is an obstacle inside its radius.

- obstacle_avoid_dist (length)
  - Default: 4*robot_radius (m)
  - This is the main parameter of the navigation method, it controls 
    the distance at which the robot slows down and begins directly 
    avoiding obstacles.  Distance is taken from the boundary of the
    robot, so the robot will navigate away from any obstacle within
    (robot_radius + obstacle_avoid_dist).  4*robot_radius is
    a good starting point, if your robot is fast or your laser sensor
    is slower than 10Hz or doesn't get reliable readings in the
    environment you may want to increase this to 6x or 8x.

- min_gap_width (length)
  - Default: 2.2*robot_radius (m)
  - The minimum width of gaps the robot will consider navigating.  
    Should not be set less than 2*robot_radius, set a bit larger to
    avoid really tight passages.

- max_speed (length/sec)
  - Default: 0.3 m/sec
  - Maximum translational speed of the robot.  Actual speed at any 
    instant is reduced based on proxmity to obstacles, rotational 
    speed, proximity to goal location.  Maximum safe speed depends on
    the refresh rate and reliability of your laser sensor, and on the
    choice of obstacle_avoid_dist.  Speeds up to 0.5 m/sec are safe for
    a Hokuyo URG-04LX and other default settings so long as the 
    environment does not have many reflective surfaces or narrow 
    obstacles.

- max_turn_rate (angle/sec)
  - Default: 60.0 deg/sec
  - Maximum rotational speed of the robot.

- goal_tol (tuple: [length angle])
  - Default: [robot_radius 30.0] (m deg)
  - Respectively, translational and rotational goal tolerance.  When the
    robot is within these bounds of the current target pose, it will
    stop.  If rotational pose is unimportant, set to 360.

@par Example

@verbatim
driver
(
  name "snd"
  provides ["position2d:1"]
  requires ["output:::position2d:0" "input:::position2d:0" "laser:0"]

  robot_radius 0.25
  obstacle_avoid_dist 1.0
  min_gap_width 0.55

  max_speed 0.4
  max_turn_rate 60

  goal_tol [.3 360]
)
@endverbatim

@author Joseph W. Durham (underlying algorithm), Luca Invernizzi (driver integration)

*/
/** @} */


#include <string.h>
#include <iostream>
#include <pthread.h>
#include <vector>
#include <libplayercore/playercore.h>

#include "snd_driver.h"
#include "snd_algo.h"


////////////////////////////////////////////////////////////////////////////////
//                           BUILDING A SHARED OBJECT                         //
////////////////////////////////////////////////////////////////////////////////

// A factory creation function, declared outside of the class so that it
// can be invoked without any object context (alternatively, you can
// declare it static in the class).  In this function, we create and return
// (as a generic Driver*) a pointer to a new instance of this driver.
Driver* SmoothND_Init(ConfigFile* cf, int section)
{
  // Create and return a new instance of this driver
  return ((Driver*) (new SmoothND_Proxy(cf, section)));
}

// A driver registration function, again declared outside of the class so
// that it can be invoked without object context.  In this function, we add
// the driver into the given driver table, indicating which interface the
// driver can support and how to create a driver instance.
void SmoothND_Register(DriverTable* table)
{
  table->AddDriver("snd", SmoothND_Init);
}

// Extra stuff for building a shared object.
/* need the extern to avoid C++ name-mangling  */
extern "C"
{
  int player_driver_init(DriverTable* table)
  {
    SmoothND_Register(table);
    return(0);
  }
}


////////////////////////////////////////////////////////////////////////////////
// Constructor.  Retrieve options from the configuration file and do any
// pre-Setup() setup.
SmoothND::SmoothND(ConfigFile* cf, int section) 
  : Driver(cf, section)
                                                  
{
    PLAYER_MSG0(1,"INITIALIZING INTERFACE ...");
    //Failsafe, shouldn't be needed
    this->goalX=0.0;
    this->goalY=0.0;
    this->data_odometry_ready=0;
    this->data_laser_ready=0;
    this->first_goal_has_been_set_to_init_position = 0;
  // Create my position interface
  if (cf->ReadDeviceAddr(&(this->m_position_addr), section, 
                         "provides", PLAYER_POSITION2D_CODE, -1, NULL) != 0)
  {
    this->SetError(-1);
    PLAYER_MSG0(1,"servi della gleba");
    return;
  }  
  if (this->AddInterface(this->m_position_addr))
  {
    this->SetError(-1);    
    PLAYER_MSG0(1,"in una stanza");
    return;
  }
  // Laser subscription address
  if (cf->ReadDeviceAddr(&(this->laser_addr), section, 
                         "requires", PLAYER_LASER_CODE, -1, NULL) != 0)
  {
    this->SetError(-1);
    return;
  }
  // Odometry subscription address
  if (cf->ReadDeviceAddr(&(this->odom_in_addr), section, 
                         "requires", PLAYER_POSITION2D_CODE, -1, "input") != 0)
  {
    this->SetError(-1);
    PLAYER_MSG0(1,"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa");
    return;
  }
  if (cf->ReadDeviceAddr(&(this->odom_out_addr), section, 
                         "requires", PLAYER_POSITION2D_CODE, -1, "output") != 0)
  {
    PLAYER_MSG0(1,"aaaaaaaaaaaaaaaaaaaaaaaBBBBBBBBBBBBBBBBBBBBaaaaaaaaaaaaa");
    this->SetError(-1);
    return;
  }
    PLAYER_MSG0(1,"INTERFACE INITIALIZED");
    
    
    robot_radius = cf->ReadLength(section, "robot_radius", 0.25);
    min_gap_width = cf->ReadLength(section, "min_gap_width", 2.2*robot_radius);
    min_gap_width = std::max(min_gap_width,2*robot_radius);
    obstacle_avoid_dist = cf->ReadLength(section, "obstacle_avoid_dist", 4*robot_radius);
    max_speed = cf->ReadLength(section, "max_speed", 0.3);
    max_turn_rate = DTOR(cf->ReadLength(section, "max_turn_rate", 60.0));
    goal_position_tol = cf->ReadTupleLength(section, "goal_tol", 0, robot_radius);
    goal_angle_tol = DTOR(cf->ReadTupleLength(section, "goal_tol", 1, 30.0));
    
    drive_mode = DRIVE_DIFF;
    if( cf->ReadInt(section, "drive_mode", 0) == 1 )
    {
        drive_mode = DRIVE_OMNI;
    }
        
}

////////////////////////////////////////////////////////////////////////////////
// Set up the device.  Return 0 if things go well, and -1 otherwise.


int SmoothND::Odometry_Setup()
{
    PLAYER_MSG0(3,"SETTING UP THE ODOMETRY  1...");
  // Subscribe to the odometry device
  if(!(this->odom_in_dev = deviceTable->GetDevice(this->odom_in_addr)))
  {
    PLAYER_ERROR("unable to locate suitable odometry device");
    return(-1);
  }
    PLAYER_MSG0(3,"SETTING UP THE ODOMETRY 2 ...");
  if(this->odom_in_dev->Subscribe(this->InQueue) != 0)
  {
    PLAYER_ERROR("unable to subscribe to odometry device");
    return(-1);
  }
    PLAYER_MSG0(3,"SETTING UP THE ODOMETRY 3 ...");
  if(!(this->odom_out_dev = deviceTable->GetDevice(this->odom_out_addr)))
  {
    PLAYER_ERROR("unable to locate suitable odometry device");
    return(-1);
  }
    PLAYER_MSG0(3,"SETTING UP THE ODOMETRY 4 ...");
  if(this->odom_out_dev->Subscribe(this->InQueue) != 0)
  {
    PLAYER_ERROR("unable to subscribe to odometry device");
    return(-1);
  }
  else{ 
      PLAYER_MSG0(0, "ok");
  }
//  sleep(3);
  PLAYER_MSG0(3,"SETTING UP THE ODOMETRY 5 ...");

//  // Get the odometry geometry
//  Message* msg;
//  if(!(msg = this->odom_out_dev->Request(this->InQueue, 
//                                  PLAYER_MSGTYPE_REQ,
//                                  PLAYER_POSITION2D_REQ_GET_GEOM,
//                                  NULL, 0, NULL,false)))
//  {
//    PLAYER_ERROR("failed to get odometry geometry");
//    if(msg)
//      delete msg;
//    return(-1);
//  }
//PLAYER_MSG0(3,"SETTING UP THE ODOMETRY 6 ...");
//  
//  player_position2d_geom_t* geom = (player_position2d_geom_t*)msg->GetPayload();
//  this->robot_geom = *geom;
//  printf("robot geom: %.3f %.3f %.3f %.3f %.3f\n",
//         this->robot_geom.size.sl,
//         this->robot_geom.size.sw,
//         this->robot_geom.pose.px,
//         this->robot_geom.pose.py,
//         RTOD(this->robot_geom.pose.pyaw));
//  delete msg;
  return 0;
}


int SmoothND::Laser_Setup()
{
  // Subscribe to the laser device
    PLAYER_MSG0(3,"SETTING UP THE LASER 1 ...");
  if(!(this->laser_dev = deviceTable->GetDevice(this->laser_addr)))
  {
    PLAYER_ERROR("unable to locate suitable laser device");
    return(-1);
  }
    PLAYER_MSG0(3,"SETTING UP THE LASER 2 ...");
  if(this->laser_dev->Subscribe(this->InQueue) != 0)
  {
      PLAYER_ERROR("unable to subscribe to laser device");
      return(-1);
  }
  PLAYER_MSG0(3,"SETTING UP THE LASER 3 ...");




//  player_laser_geom_t* cfg;
//  Message* msg;
//  // Get the laser pose
//  if(!(msg = this->laser_dev->Request(this->InQueue, 
//                  PLAYER_MSGTYPE_REQ,
//                  PLAYER_LASER_REQ_GET_GEOM,
//                  NULL, 0, NULL,false)))
//  {
//      PLAYER_ERROR("failed to get laser geometry");
//      return(-1);
//  }
//
//  // Store the laser pose
//  cfg = (player_laser_geom_t*)msg->GetPayload();
//  // this->laser_pose = cfg->pose;
//  delete msg;
//

  return 0;
}


int SmoothND::Setup()
{   
    PLAYER_MSG0(1,"SETTING UP THE DRIVER ...");
    
    if (this->Odometry_Setup()!=0 ) return -1;
    if (this->Laser_Setup() !=0   ) return -1;

  // Here you do whatever else is necessary to setup the device, like open and
  // configure a serial port.
    
    PLAYER_MSG0(1,"DRIVER READY");

  // Start the device thread; spawns a new thread and executes
  // SmoothND::Main(), which contains the main loop for the driver.
  this->StartThread();

  return(0);
}


////////////////////////////////////////////////////////////////////////////////
// Shutdown the device
int SmoothND::Shutdown()
{

  // Stop and join the driver thread
  this->StopThread();

  // Unsubscribe from the laser
  this->laser_dev->Unsubscribe(this->InQueue);

  // Unsubscribe from the odometry
 // this->PutPositionCmd(0.0,0.0);
  this->odom_in_dev->Unsubscribe(this->InQueue);
  this->odom_out_dev->Unsubscribe(this->InQueue);

  // Here you would shut the device down by, for example, closing a
  // serial port.


  return(0);
}


////////////////////////////////////////////////////////////////////////////////
// Main function for device thread
void SmoothND::Main() 
{
    pthread_attr_t thread_attr;
    pthread_mutexattr_t goal_mutex_attr, data_mutex_attr;
    pthread_condattr_t goal_changed_cond_attr, data_changed_cond_attr;

    pthread_condattr_init(&goal_changed_cond_attr);
    pthread_cond_init(&(this->goal_changed_cond), &goal_changed_cond_attr);
    pthread_condattr_destroy(&goal_changed_cond_attr);

    pthread_mutexattr_init(&goal_mutex_attr);
    pthread_mutex_init(&(this->goal_mutex), &goal_mutex_attr);
    pthread_mutexattr_destroy(&goal_mutex_attr);

    pthread_condattr_init(&data_changed_cond_attr);
    pthread_cond_init(&(this->data_changed_cond), &data_changed_cond_attr);
    pthread_condattr_destroy(&data_changed_cond_attr);

    pthread_mutexattr_init(&data_mutex_attr);
    pthread_mutex_init(&(this->data_mutex), &data_mutex_attr);
    pthread_mutexattr_destroy(&data_mutex_attr);


    // Wait for the first goal to be set to the initial position
    while (first_goal_has_been_set_to_init_position==0){
        PLAYER_MSG0(2,".");
        this->Wait();
        this->ProcessMessages();
    }
    pthread_attr_init(&thread_attr);
    pthread_attr_setinheritsched(&thread_attr, PTHREAD_EXPLICIT_SCHED);
    pthread_attr_setschedpolicy(&thread_attr, SCHED_RR);


    if ( pthread_create( &(this->algorithm_thread), NULL, start_algorithm, (void*)this ) ) {
        cerr << "error creating thread" << endl;
    }
    PLAYER_MSG0(2,"GOING");
    pthread_attr_destroy(&thread_attr);


  // The main loop; interact with the device here
  for(;;)
  {
    // test if we are supposed to cancel
    pthread_testcancel();

    // Process incoming messages.  Calls ProcessMessage() on each pending
    // message.
    this->ProcessMessages();

    // Sleep (or you might, for example, block on a read() instead)
    usleep(50000);
  }
  return;
}

////////////////////////////////////////////////////////////////////////////////

void SmoothND::SetSpeedCmd(player_position2d_cmd_vel_t cmd)
{
//  cmd.state = 1;
  // cmd.type = 0;
  this->odom_out_dev->PutMsg(this->InQueue,
                             PLAYER_MSGTYPE_CMD,
                             PLAYER_POSITION2D_CMD_VEL,
                             (void*)&cmd,sizeof(cmd),NULL);
}


////////////////////////////////////////////////////////////////////////////////


int SmoothND::ProcessMessage(QueuePointer & resp_queue, 
                                player_msghdr * hdr, 
                                void * data)
{

  // Handle new data from the laser
    if(Message::MatchMessage(hdr, PLAYER_MSGTYPE_DATA, PLAYER_LASER_DATA_SCAN, 
                this->laser_addr))
    {

          //PLAYER_MSG0(2,"INCOMING LASER SCAN");
        pthread_mutex_lock(&(this->data_mutex));
        this->laser__ranges_count = ((player_laser_data_t*) data)->ranges_count  ;
        this->laser__resolution=((player_laser_data_t*) data)->resolution;
        this->laser__max_range=((player_laser_data_t*) data)->max_range;
        this->laser__ranges.clear();
        for (unsigned int index=0;index<this->laser__ranges_count;index ++) 
        {
            this->laser__ranges.push_back(((player_laser_data_t*) data)->ranges[index]) ;
        }
        this->data_laser_ready=1;
        pthread_mutex_unlock(&(this->data_mutex));
        this->SignalRead();
        return(0);
    }

    if(Message::MatchMessage(hdr, PLAYER_MSGTYPE_DATA, 
                           PLAYER_POSITION2D_DATA_STATE, 
                           this->odom_in_addr))
  {
    pthread_mutex_lock(&(this->data_mutex));
    this->odom_pose= ((player_position2d_data_t*) data)->pos;
    if (this->first_goal_has_been_set_to_init_position ==0) {
        this->SignalNextGoal(this->odom_pose.px,this->odom_pose.py,this->odom_pose.pa);
        this->first_goal_has_been_set_to_init_position=1;
    }
    hdr->addr = this->device_addr;
    this->data_odometry_ready=1;
    pthread_mutex_unlock(&(this->data_mutex));
    this->Publish(hdr,data);
    PLAYER_MSG3(2, "Here I am: (%.3f %.3f %.3f)",this->odom_pose.px,this->odom_pose.py,this->odom_pose.pa);
    this->SignalRead();
    return(0);
  }


  
  if(Message::MatchMessage(hdr, PLAYER_MSGTYPE_DATA, 
                           PLAYER_POSITION2D_DATA_STATE, 
                           this->odom_out_addr)) {
    //hdr->addr = this->device_addr;
    //this->Publish(hdr,data);
    return(0);
  }

  if(Message::MatchMessage(hdr, PLAYER_MSGTYPE_REQ, -1, this->device_addr))
  {
    Message* msg;
    if(!(msg = this->odom_out_dev->Request(this->InQueue,
                                       hdr->type,
                                       hdr->subtype,
                                       (void*)data, 
                                       hdr->size, 
                                       &hdr->timestamp)))
    {
      PLAYER_WARN1("failed to forward config request with subtype: %d\n", hdr->subtype);
      return(-1);
    }
    player_msghdr_t* rephdr = msg->GetHeader();
    void* repdata = msg->GetPayload();
    // Copy in our address and forward the response
    rephdr->addr = this->device_addr;
    this->Publish(resp_queue, rephdr, repdata);
    delete msg;
    return(0);
  }

  if(Message::MatchMessage(hdr, PLAYER_MSGTYPE_CMD, 
                   PLAYER_POSITION2D_CMD_POS,
                                this->device_addr))
  {
    player_position2d_cmd_pos_t* cmd = (player_position2d_cmd_pos_t*)data;
//    PLAYER_MSG3(2, "New goal (position): (%.3f %.3f %.3f)",
//              cmd->pos.px, 
//              cmd->pos.py, 
//              RTOD(cmd->pos.pa));
    this->SignalNextGoal(cmd->pos.px,cmd->pos.py,cmd->pos.pa);
    return 0;
    
  }

  if(Message::MatchMessage(hdr, PLAYER_MSGTYPE_CMD, 
                   PLAYER_POSITION2D_CMD_VEL,
                                this->device_addr))
  {
    player_position2d_cmd_vel_t* cmd = (player_position2d_cmd_vel_t*)data;
    //PLAYER_MSG3(1, "New goal (speed): (%.3f %.3f %.3f)",
    //          cmd->vel.px, 
    //          cmd->vel.py, 
    //          RTOD(cmd->vel.pa));
    this->SetSpeedCmd(*cmd);
/*    if( cmd->vel.px < .05 && cmd->vel.pa < DTOR(3.0) )
    {
        this->SignalNextGoal(this->odom_pose.px, 
                                this->odom_pose.py,
                                this->odom_pose.pa);
    }
    else
    {
        double angle = ((cmd->vel.px < 0) ? M_PI : 0) + 2*cmd->vel.pa; 
        this->SignalNextGoal(this->odom_pose.px + 10000, 
                                this->odom_pose.py + 10000,
                                angle); 
    }     
*/
    //PLAYER_WARN1("cmd->state:%d\n",cmd->state);
    return 0;
    
  }
  // Tell the caller that you don't know how to handle this message
    PLAYER_MSG0(1,"PUPPA!");
  return(-1);
}



//////////////////////////////////////////////////////7

double SmoothND_Proxy::GetScanRes() 
{

    PLAYER_MSG0(2,"GetScanRes");
    pthread_mutex_lock(&(this->data_mutex));
    double temp=this->laser__resolution;
    pthread_mutex_unlock(&(this->data_mutex));
    return temp;
}


double SmoothND_Proxy::GetMaxRange() 
{
    PLAYER_MSG0(2,"GetMaxRange");
    pthread_mutex_lock(&(this->data_mutex));
    double temp=this->laser__max_range;
    pthread_mutex_unlock(&(this->data_mutex));
    return temp;
}


uint32_t SmoothND_Proxy::GetCount()
{
    PLAYER_MSG0(2,"GetCount");
    pthread_mutex_lock(&(this->data_mutex));
    int temp=this->laser__ranges_count;
    pthread_mutex_unlock(&(this->data_mutex));
    return temp;
}

double SmoothND_Proxy::range(const int index)
{
    pthread_mutex_lock(&(this->data_mutex));
    double temp= this->laser__ranges[index];
    pthread_mutex_unlock(&(this->data_mutex));
    return temp;
}





void   SmoothND_Proxy::SetMotorEnable(int turnkey)
{
    PLAYER_MSG0(1,"MOTOR_ENABLE -- not implemented!");
    
}
void   SmoothND_Proxy::SetOdometry(double position_x0,
                                   double position_y0,
                                   double position_alpha0)
{
    PLAYER_MSG0(1,"SET_ODOMETRY -- not implemented!");

}
double SmoothND_Proxy::GetXPos() 
{
    pthread_mutex_lock(&(this->data_mutex));
    double temp= this->odom_pose.px;
    pthread_mutex_unlock(&(this->data_mutex));
    return temp;
}
double SmoothND_Proxy::GetYPos()
{
    pthread_mutex_lock(&(this->data_mutex));
    double temp= this->odom_pose.py;
    pthread_mutex_unlock(&(this->data_mutex));
    return temp;
}
double SmoothND_Proxy::GetYaw()
{
    pthread_mutex_lock(&(this->data_mutex));
    double temp= this->odom_pose.pa;
    pthread_mutex_unlock(&(this->data_mutex));
    return temp;
}
void   SmoothND_Proxy::RequestGeom() 
{
    //done automatically
}
void   SmoothND_Proxy::SetSpeed(double velocity_modulus,
                                double velocity_angle) 
{
    player_position2d_cmd_vel_t cmd;
    cmd.vel.px=velocity_modulus;
    cmd.vel.py=0.0;
    cmd.vel.pa=velocity_angle;
    this->SetSpeedCmd(cmd);

}

void   SmoothND_Proxy::SetSpeed(double velocity_x,
                                double velocity_y,
                                double velocity_angle) 
{
    player_position2d_cmd_vel_t cmd;
    cmd.vel.px=velocity_x;
    cmd.vel.py=velocity_y;
    cmd.vel.pa=velocity_angle;
    this->SetSpeedCmd(cmd);

}


void unlock_mutex(void *m) 
{
      pthread_mutex_unlock((pthread_mutex_t *)m);
}

void SmoothND::WaitForNextGoal() 
{
      pthread_mutex_lock(&(this->goal_mutex));
      pthread_cleanup_push(unlock_mutex,(void *)&(this->goal_mutex));
      pthread_cond_wait(&(this->goal_changed_cond), &(this->goal_mutex));
      pthread_cleanup_pop(0);
      pthread_mutex_unlock(&(this->goal_mutex));
}

void SmoothND::SignalNextGoal(double goalX,double goalY,double goalA) 
{
      pthread_mutex_lock(&(this->goal_mutex));
      this->goalX=goalX;
      this->goalY=goalY;
      this->goalA=goalA;
      pthread_cond_signal(&(this->goal_changed_cond));
      pthread_mutex_unlock(&(this->goal_mutex));
      PLAYER_MSG0(2,"SIGNAL Goal");
}
void SmoothND::Read() 
{
      PLAYER_MSG0(2,"WAiting for data");
      pthread_mutex_lock(&(this->data_mutex));
      pthread_cleanup_push(unlock_mutex,(void *)&(this->data_mutex));
      pthread_cond_wait(&(this->data_changed_cond), &(this->data_mutex));
      pthread_cleanup_pop(0);
      pthread_mutex_unlock(&(this->data_mutex));
}

void SmoothND::SignalRead() 
{
    if (this->data_odometry_ready==1 && this->data_laser_ready==1)
    {
        PLAYER_MSG0(2,"SIGNAL DATA READY");
        this->data_odometry_ready=0;
        this->data_laser_ready=0;
        pthread_mutex_lock(&(this->data_mutex));
        pthread_cond_signal(&(this->data_changed_cond));
        pthread_mutex_unlock(&(this->data_mutex));
    }
}
